import sys
import secrets
from PyQt5.QtWidgets import (
    QApplication, QMainWindow, QWidget, QVBoxLayout, QHBoxLayout,
    QPushButton, QLabel, QFrame, QSizePolicy, QStyle
)
from PyQt5.QtCore import Qt, QTimer, QSize
from PyQt5.QtGui import QFont
from phone_client import PhoneClient
from waveform_widget import WaveformWidget
from phone_state import PhoneState
from session import NoiseXKSession

class PhoneUI(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Enhanced Dual Phone Interface")
        self.setGeometry(100, 100, 900, 750)
        self.setStyleSheet("""
            QMainWindow { background-color: #333333; }
            QLabel { color: #E0E0E0; font-size: 14px; }
            QPushButton {
                background-color: #0078D4; color: white; border: none;
                padding: 10px 15px; border-radius: 5px; font-size: 14px;
                min-height: 30px;
            }
            QPushButton:hover { background-color: #005A9E; }
            QPushButton:pressed { background-color: #003C6B; }
            QPushButton#settingsButton { background-color: #555555; }
            QPushButton#settingsButton:hover { background-color: #777777; }
            QFrame#phoneDisplay {
                background-color: #1E1E1E; border: 2px solid #0078D4;
                border-radius: 10px;
            }
            QLabel#phoneTitleLabel {
                font-size: 18px; font-weight: bold; padding-bottom: 5px;
                color: #FFFFFF;
            }
            QLabel#mainTitleLabel {
                font-size: 24px; font-weight: bold; color: #00A2E8;
                padding: 15px;
            }
            QWidget#phoneWidget {
                border: 1px solid #4A4A4A; border-radius: 8px;
                padding: 10px; background-color: #3A3A3A;
            }
        """)

        # Main widget and layout
        main_widget = QWidget()
        self.setCentralWidget(main_widget)
        main_layout = QVBoxLayout()
        main_layout.setSpacing(20)
        main_layout.setContentsMargins(20, 20, 20, 20)
        main_layout.setAlignment(Qt.AlignCenter)
        main_widget.setLayout(main_layout)

        # App Title
        app_title_label = QLabel("Dual Phone Control Panel")
        app_title_label.setObjectName("mainTitleLabel")
        app_title_label.setAlignment(Qt.AlignCenter)
        main_layout.addWidget(app_title_label)

        # Phone displays layout
        phone_controls_layout = QHBoxLayout()
        phone_controls_layout.setSpacing(50)
        phone_controls_layout.setAlignment(Qt.AlignCenter)
        main_layout.addLayout(phone_controls_layout)

        # Initialize phones
        self.phones = []
        self.handshake_done_count = 0
        for i in range(2):
            client = PhoneClient("localhost", 12345, i)
            client.data_received.connect(lambda data, cid=i: self.update_waveform(cid, data))
            client.state_changed.connect(lambda state, num, cid=i: self.set_phone_state(cid, self.map_state(state), num))
            client.start()

            # Generate keypair for each phone
            keypair = NoiseXKSession.generate_keypair()

            phone_container_widget, phone_display_frame, phone_button, waveform_widget, phone_status_label = self._create_phone_ui(
                f"Phone {i+1}", lambda checked, phone_id=i: self.phone_action(phone_id)
            )
            self.phones.append({
                'id': i,
                'client': client,
                'state': PhoneState.IDLE,
                'button': phone_button,
                'waveform': waveform_widget,
                'number': "123-4567" if i == 0 else "987-6543",
                'audio_timer': None,
                'status_label': phone_status_label,
                'keypair': keypair,
                'public_key': keypair.public,
                'is_initiator': False
            })
            phone_controls_layout.addWidget(phone_container_widget)

        # Share public key between phones
        self.phones[0]['peer_public_key'] = self.phones[1]['public_key']
        self.phones[1]['peer_public_key'] = self.phones[0]['public_key']

        # Spacer
        main_layout.addStretch(1)

        # Settings Button
        self.settings_button = QPushButton("Settings")
        self.settings_button.setObjectName("settingsButton")
        self.settings_button.setFixedWidth(180)
        self.settings_button.setIcon(self.style().standardIcon(QStyle.SP_FileDialogDetailedView))
        self.settings_button.setIconSize(QSize(20, 20))
        self.settings_button.clicked.connect(self.settings_action)
        settings_layout = QHBoxLayout()
        settings_layout.addStretch()
        settings_layout.addWidget(self.settings_button)
        settings_layout.addStretch()
        main_layout.addLayout(settings_layout)

        # Initialize button states
        for phone in self.phones:
            self._update_phone_button_ui(phone['button'], phone['status_label'], phone['state'])

    def _create_phone_ui(self, title, action_slot):
        phone_container_widget = QWidget()
        phone_container_widget.setObjectName("phoneWidget")
        phone_layout = QVBoxLayout()
        phone_layout.setAlignment(Qt.AlignCenter)
        phone_layout.setSpacing(15)
        phone_container_widget.setLayout(phone_layout)

        phone_title_label = QLabel(title)
        phone_title_label.setObjectName("phoneTitleLabel")
        phone_title_label.setAlignment(Qt.AlignCenter)
        phone_layout.addWidget(phone_title_label)

        phone_display_frame = QFrame()
        phone_display_frame.setObjectName("phoneDisplay")
        phone_display_frame.setFixedSize(250, 350)
        phone_display_frame.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

        display_content_layout = QVBoxLayout(phone_display_frame)
        display_content_layout.setAlignment(Qt.AlignCenter)
        phone_status_label = QLabel("Idle")
        phone_status_label.setAlignment(Qt.AlignCenter)
        phone_status_label.setFont(QFont("Arial", 16))
        display_content_layout.addWidget(phone_status_label)
        phone_layout.addWidget(phone_display_frame, alignment=Qt.AlignCenter)

        phone_button = QPushButton()
        phone_button.setFixedWidth(120)
        phone_button.setIconSize(QSize(20, 20))
        phone_button.clicked.connect(action_slot)
        phone_layout.addWidget(phone_button, alignment=Qt.AlignCenter)

        waveform_label = QLabel(f"{title} Audio")
        waveform_label.setAlignment(Qt.AlignCenter)
        waveform_label.setStyleSheet("font-size: 14px; color: #E0E0E0;")
        phone_layout.addWidget(waveform_label)
        waveform_widget = WaveformWidget(dynamic=False)
        phone_layout.addWidget(waveform_widget, alignment=Qt.AlignCenter)

        return phone_container_widget, phone_display_frame, phone_button, waveform_widget, phone_status_label

    def _update_phone_button_ui(self, button, status_label, state, phone_number=""):
        if state == PhoneState.IDLE:
            button.setText("Call")
            button.setIcon(self.style().standardIcon(QStyle.SP_MediaPlay))
            status_label.setText("Idle")
            button.setStyleSheet("background-color: #0078D4;")
        elif state == PhoneState.CALLING:
            button.setText("Cancel")
            button.setIcon(self.style().standardIcon(QStyle.SP_MediaStop))
            status_label.setText(f"Calling {phone_number}...")
            button.setStyleSheet("background-color: #E81123;")
        elif state == PhoneState.IN_CALL:
            button.setText("Hang Up")
            button.setIcon(self.style().standardIcon(QStyle.SP_DialogCancelButton))
            status_label.setText(f"In Call with {phone_number}")
            button.setStyleSheet("background-color: #E81123;")
        elif state == PhoneState.RINGING:
            button.setText("Answer")
            button.setIcon(self.style().standardIcon(QStyle.SP_DialogApplyButton))
            status_label.setText(f"Incoming Call from {phone_number}")
            button.setStyleSheet("background-color: #107C10;")

    def phone_action(self, phone_id):
        phone = self.phones[phone_id]
        other_phone = self.phones[1 - phone_id]
        print(f"Phone {phone_id + 1} Action, current state: {phone['state']}, is_initiator: {phone['is_initiator']}")
        
        if phone['state'] == PhoneState.IDLE:
            # Initiate a call
            phone['state'] = PhoneState.CALLING
            other_phone['state'] = PhoneState.RINGING
            # Set init/resp
            phone['is_initiator'] = True
            other_phone['is_initiator'] = False
            self._update_phone_button_ui(phone['button'], phone['status_label'], phone['state'], other_phone['number'])
            self._update_phone_button_ui(other_phone['button'], other_phone['status_label'], other_phone['state'], phone['number'])
            phone['client'].send("RINGING")
        
        elif phone['state'] == PhoneState.RINGING:
            # Answer the call
            phone['state'] = PhoneState.IN_CALL
            other_phone['state'] = PhoneState.IN_CALL
            self._update_phone_button_ui(phone['button'], phone['status_label'], phone['state'], other_phone['number'])
            self._update_phone_button_ui(other_phone['button'], other_phone['status_label'], other_phone['state'], phone['number'])
            phone['client'].send("IN_CALL")
        
        elif phone['state'] == PhoneState.IN_CALL or phone['state'] == PhoneState.CALLING:
            # Hang up or cancel
            if not phone['client'].handshake_in_progress and phone['state'] != PhoneState.CALLING:
                phone['state'] = PhoneState.IDLE
                other_phone['state'] = PhoneState.IDLE
                self._update_phone_button_ui(phone['button'], phone['status_label'], phone['state'], "")
                self._update_phone_button_ui(other_phone['button'], other_phone['status_label'], other_phone['state'], "")
                phone['client'].send("CALL_END")
                # Stop audio timers for both phones
                for p in [phone, other_phone]:
                    if p['audio_timer']:
                        p['audio_timer'].stop()
            else:
                print(f"Phone {phone_id + 1} cannot hang up during handshake or call setup")

    def start_audio(self, client_id):
        """Start audio timer after both clients send HANDSHAKE_DONE."""
        self.handshake_done_count += 1
        print(f"HANDSHAKE_DONE received for client {client_id}, count: {self.handshake_done_count}")
        if self.handshake_done_count == 2:
            for phone in self.phones:
                if phone['state'] == PhoneState.IN_CALL:
                    if not phone['audio_timer'] or not phone['audio_timer'].isActive():
                        phone['audio_timer'] = QTimer(self)
                        phone['audio_timer'].timeout.connect(lambda pid=phone['id']: self.send_audio(pid))
                        phone['audio_timer'].start(100)  # 100ms for smoother updates
            self.handshake_done_count = 0

    def send_audio(self, phone_id):
        phone = self.phones[phone_id]
        if phone['state'] == PhoneState.IN_CALL and phone['client'].session and phone['client'].sock:
            # Generate mock 16-byte audio data
            mock_audio = secrets.token_bytes(16)
            try:
                # Encrypt with Noise session, send over socket
                phone['client'].session.send(phone['client'].sock, mock_audio)
                print(f"Client {phone_id} sent encrypted audio packet, length=32")
            except Exception as e:
                print(f"Client {phone_id} failed to send audio: {e}")

    def update_waveform(self, client_id, data):
        print(f"Updating waveform for client_id {client_id}, data_length={len(data)}")
        waveform = self.phones[client_id]['waveform']
        waveform.set_data(data)

    def map_state(self, state_str):
        if state_str == "RINGING":
            return PhoneState.RINGING
        elif state_str in ["CALL_END", "CALL_DROPPED"]:
            return PhoneState.IDLE
        elif state_str == "IN_CALL":
            return PhoneState.IN_CALL
        elif state_str == "HANDSHAKE":
            return PhoneState.IN_CALL  # Stay in IN_CALL, trigger handshake
        elif state_str == "HANDSHAKE_DONE":
            return PhoneState.IN_CALL  # Stay in IN_CALL, start audio
        return PhoneState.IDLE

    def set_phone_state(self, client_id, state, number=""):
        phone = self.phones[client_id]
        other_phone = self.phones[1 - client_id]
        print(f"Setting state for Phone {client_id + 1}: {state}, number: {number}, is_initiator: {phone['is_initiator']}")
        phone['state'] = state
        if state == PhoneState.RINGING:
            self._update_phone_button_ui(phone['button'], phone['status_label'], state, other_phone['number'])
        elif state == PhoneState.IN_CALL:
            print(f"Phone {client_id + 1} confirmed in IN_CALL state")
            self._update_phone_button_ui(phone['button'], phone['status_label'], state, other_phone['number'])
            if number == "IN_CALL" and phone['is_initiator']:
                # Initiator starts handshake after receiving IN_CALL
                print(f"Phone {client_id + 1} (initiator) starting handshake")
                phone['client'].send("HANDSHAKE")
                phone['client'].start_handshake(initiator=True, keypair=phone['keypair'], peer_pubkey=other_phone['public_key'])
            elif number == "HANDSHAKE" and not phone['is_initiator']:
                # Responder starts handshake after receiving HANDSHAKE
                print(f"Phone {client_id + 1} (responder) starting handshake")
                phone['client'].start_handshake(initiator=False, keypair=phone['keypair'], peer_pubkey=other_phone['public_key'])
            elif number == "HANDSHAKE_DONE":
                # Start audio after HANDSHAKE_DONE
                self.start_audio(client_id)
        else:
            # Handle disconnect gracefully
            self._update_phone_button_ui(phone['button'], phone['status_label'], state, "")
            if state == PhoneState.IDLE and number == "CALL_END":
                print(f"Phone {client_id + 1} resetting due to disconnect")
        if state == PhoneState.IDLE and phone['audio_timer']:
            phone['audio_timer'].stop()

    def settings_action(self):
        print("Settings clicked")

    def closeEvent(self, event):
        for phone in self.phones:
            phone['client'].stop()
        event.accept()

if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = PhoneUI()
    window.show()
    sys.exit(app.exec_())