import socket
import time
import select
from PyQt5.QtCore import QThread, pyqtSignal
from queue import Queue
from session import NoiseXKSession

class PhoneClient(QThread):
    data_received = pyqtSignal(bytes, int)  # Include client_id
    state_changed = pyqtSignal(str, str, int)  # Include client_id

    def __init__(self, host, port, client_id):
        super().__init__()
        self.host = host
        self.port = port
        self.client_id = client_id
        self.sock = None
        self.running = True
        self.command_queue = Queue()
        self.initiator = None
        self.keypair = None
        self.peer_pubkey = None
        self.session = None
        self.handshake_in_progress = False
        self.handshake_start_time = None
        self.call_active = False  # Track active call after HANDSHAKE_DONE

    def connect_socket(self):
        """Attempt to connect to the server with retries."""
        retries = 3
        for attempt in range(retries):
            try:
                self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                self.sock.settimeout(120)  # 120s for socket operations
                self.sock.connect((self.host, self.port))
                print(f"Client {self.client_id} connected to {self.host}:{self.port}")
                return True
            except Exception as e:
                print(f"Client {self.client_id} connection attempt {attempt + 1} failed: {e}")
                if attempt < retries - 1:
                    time.sleep(1)  # Wait before retrying
                self.sock = None
        return False

    def run(self):
        while self.running:
            if not self.sock:
                if not self.connect_socket():
                    print(f"Client {self.client_id} failed to connect after retries")
                    self.state_changed.emit("CALL_END", "", self.client_id)
                    break
            try:
                while self.running:
                    # print(f"Client {self.client_id} run loop iteration")
                    # Check command queue first
                    if not self.command_queue.empty():
                        print(f"Client {self.client_id} processing command queue, size: {self.command_queue.qsize()}")
                        command = self.command_queue.get()
                        if command == "handshake":
                            try:
                                print(f"Client {self.client_id} starting handshake, initiator: {self.initiator}")
                                self.session = NoiseXKSession(self.keypair, self.peer_pubkey)
                                self.session.handshake(self.sock, self.initiator)
                                print(f"Client {self.client_id} handshake complete")
                                self.send("HANDSHAKE_DONE")
                            except socket.timeout:
                                print(f"Client {self.client_id} handshake timed out")
                                self.state_changed.emit("CALL_END", "", self.client_id)
                                break
                            except Exception as e:
                                print(f"Client {self.client_id} handshake failed: {e}")
                                self.state_changed.emit("CALL_END", "", self.client_id)
                                break
                            finally:
                                self.handshake_in_progress = False
                                self.handshake_start_time = None
                    else:
                        # Check for handshake timeout
                        if self.handshake_in_progress and self.handshake_start_time:
                            if time.time() - self.handshake_start_time > 30:  # 30s handshake timeout
                                print(f"Client {self.client_id} handshake timeout after 30s")
                                self.state_changed.emit("CALL_END", "", self.client_id)
                                self.handshake_in_progress = False
                                self.handshake_start_time = None
                                break
                        # Only read socket if not in handshake
                        if not self.handshake_in_progress:
                            # Use select to check if data is available
                            readable, _, _ = select.select([self.sock], [], [], 0.01)  # 10ms timeout
                            if readable:
                                try:
                                    print(f"Client {self.client_id} attempting sock.recv")
                                    data = self.sock.recv(1024)
                                    if not data:
                                        print(f"Client {self.client_id} disconnected")
                                        self.state_changed.emit("CALL_END", "", self.client_id)
                                        break
                                    # Handle control messages (UTF-8)
                                    try:
                                        decoded_data = data.decode('utf-8').strip()
                                        print(f"Client {self.client_id} received raw: {decoded_data}")
                                        if decoded_data in ["RINGING", "CALL_END", "CALL_DROPPED", "IN_CALL", "HANDSHAKE", "HANDSHAKE_DONE"]:
                                            self.state_changed.emit(decoded_data, decoded_data, self.client_id)
                                            if decoded_data == "HANDSHAKE":
                                                self.handshake_in_progress = True  # Block further reads
                                            elif decoded_data == "HANDSHAKE_DONE":
                                                self.call_active = True  # Enable audio processing
                                        else:
                                            print(f"Client {self.client_id} ignored unexpected text message: {decoded_data}")
                                    except UnicodeDecodeError:
                                        # Handle binary data (audio packets)
                                        if self.call_active and self.session:
                                            try:
                                                print(f"Client {self.client_id} received audio packet, length={len(data)}")
                                                decrypted_data = self.session.decrypt(data)
                                                print(f"Client {self.client_id} decrypted audio packet, length={len(decrypted_data)}")
                                                self.data_received.emit(decrypted_data, self.client_id)
                                            except Exception as e:
                                                print(f"Client {self.client_id} failed to process audio packet: {e}")
                                        else:
                                            print(f"Client {self.client_id} ignored non-text message: {data.hex()}")
                                except socket.timeout:
                                    print(f"Client {self.client_id} timed out waiting for data")
                                    continue
                                except socket.error as e:
                                    print(f"Client {self.client_id} socket error: {e}")
                                    self.state_changed.emit("CALL_END", "", self.client_id)
                                    break
                                except Exception as e:
                                    print(f"Client {self.client_id} error: {e}")
                                    self.state_changed.emit("CALL_END", "", self.client_id)
                                    break
                            else:
                                # print(f"Client {self.client_id} no data available, skipping recv")
                                pass
                        else:
                            # Yield during handshake
                            self.msleep(20)  # 20ms sleep to yield CPU
                            print(f"Client {self.client_id} yielding during handshake")
                    # Short sleep to yield Qt event loop
                    self.msleep(1)  # 1ms sleep
            finally:
                if self.sock:
                    self.sock.close()
                    self.sock = None

    def send(self, message):
        if self.sock and self.running:
            try:
                if isinstance(message, str):
                    data = message.encode('utf-8')
                    self.sock.send(data)
                    print(f"Client {self.client_id} sent: {message}, length={len(data)}")
                else:
                    # Send binary data (audio)
                    self.sock.send(message)
                    print(f"Client {self.client_id} sent binary data, length={len(message)}")
            except socket.error as e:
                print(f"Client {self.client_id} send error: {e}")
                self.state_changed.emit("CALL_END", "", self.client_id)

    def stop(self):
        self.running = False
        if self.sock:
            self.sock.close()
            self.sock = None
    
    def start_handshake(self, initiator, keypair, peer_pubkey):
        """Queue the handshake command with necessary parameters."""
        self.initiator = initiator
        self.keypair = keypair
        self.peer_pubkey = peer_pubkey
        print(f"Client {self.client_id} queuing handshake, initiator: {initiator}")
        self.handshake_in_progress = True  # Block recv before handshake starts
        self.handshake_start_time = time.time()
        self.command_queue.put("handshake")